//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%
#include <fstream>
#include <algorithm>

// ResponseMatrix
#include <XSModel/Data/Detector/ResponseMatrix.h>
#include "XSstreams.h"
#include <XSModel/Data/DataUtility.h>
#include "XSsymbol.h"
#include <XSUtil/Utils/XSstream.h>
#include <XSUtil/Parse/XSparse.h>
#include <cstring>

#include "rmf.h"

// Class ResponseMatrix::ChanRangeIndicators 

// Class ResponseMatrix 

ResponseMatrix::ResponseMatrix ()
      : m_eMin(0),
        m_eMax(0),
        m_name(""),
	m_gqString(),
        m_telescope(),
        m_instrument(),
        m_channelType(),
        m_retainPosition(),
        m_convArraySize(0),
        m_detectorChannels(),
        m_chanRangeIndicator(),
        m_energyHigh(),
        m_energyLow(),
        m_binStart(),
        m_binRunLengths(),
        m_eboundsMin(),
        m_eboundsMax(),
        m_matrixValue(),
	m_normFactor(),
        m_convArray(),
        m_binResponseGroups(),
        m_channelForRun(),
        m_energyRunLengths(),
        m_energyStart(),
        m_positions()
{
}

ResponseMatrix::ResponseMatrix(const ResponseMatrix &right)
      : m_eMin(right.m_eMin),
        m_eMax(right.m_eMax),
        m_name(right.m_name),
	m_gqString(right.m_gqString),
        m_telescope(right.m_telescope),
        m_instrument(right.m_instrument),
        m_channelType(right.m_channelType),
        m_retainPosition(right.m_retainPosition),
        m_convArraySize(right.m_convArraySize),
        m_detectorChannels(right.m_detectorChannels),
        m_chanRangeIndicator(right.m_chanRangeIndicator),
        m_energyHigh(right.m_energyHigh),
        m_energyLow(right.m_energyLow),
        m_binStart(right.m_binStart),
        m_binRunLengths(right.m_binRunLengths),
        m_eboundsMin(right.m_eboundsMin),
        m_eboundsMax(right.m_eboundsMax),
        m_matrixValue(right.m_matrixValue),
	m_normFactor(right.m_normFactor),
        m_convArray(),
        m_binResponseGroups(right.m_binResponseGroups),
        m_channelForRun(right.m_channelForRun),
        m_energyRunLengths(right.m_energyRunLengths),
        m_energyStart(right.m_energyStart),
        m_positions(right.m_positions)
{
  m_convArray.reset(new Real[m_convArraySize]);
  memcpy(m_convArray.get(),right.m_convArray.get(),m_convArraySize);
}

ResponseMatrix::ResponseMatrix (const ResponseMatrix::ChanRangeIndicators& chanLimits, size_t nE, size_t nC, const string& id, const CodeContainer& gqString, bool retainPositionArray)
      : m_eMin(0),
        m_eMax(0),
        m_name(id),
	m_gqString(gqString),
        m_telescope(),
        m_instrument(),
        m_channelType(),
        m_retainPosition(retainPositionArray),
        m_convArraySize(0),
        m_detectorChannels(nC),
        m_chanRangeIndicator(chanLimits),
        m_energyHigh(nE),
        m_energyLow(nE),
        m_binStart(nE),
        m_binRunLengths(nE),
        m_eboundsMin(nC),
        m_eboundsMax(nC),
        m_matrixValue(nE),
	m_normFactor(),
        m_convArray(),
        m_binResponseGroups(nE),
        m_channelForRun(),
        m_energyRunLengths(),
        m_energyStart(),
        m_positions()
{
}



ResponseMatrix::~ResponseMatrix()
{
}



int ResponseMatrix::operator==(const ResponseMatrix &right) const
{
  return compare(right);
}

int ResponseMatrix::operator!=(const ResponseMatrix &right) const
{
  return !compare(right);
}


ResponseMatrix* ResponseMatrix::clone () const
{

  return new ResponseMatrix(*this);
}

bool ResponseMatrix::compare (const ResponseMatrix& right) const
{
  if (m_name != right.m_name) return false;
  if (m_gqString != right.m_gqString) return false;
  return true;
}

// read from a file
void ResponseMatrix::read(const string& fileName, const ResponseMatrix::ChanRangeIndicators& chanLimits, const int spectrumDetchans, const CodeContainer& gqString, bool retainPositionArray)
{
  int rmfNum = 1;
  string openFileName = fileName;
  // fileName may have a {n} suffix so check for this and split it off
  size_t openParenPos = fileName.find_first_of('{');
  if ( openParenPos  != string::npos ) {
    openFileName = fileName.substr(0,openParenPos);
    size_t closeParenPos = fileName.find_first_of('}');
    if ( !SPstring2Integer(fileName.substr(openParenPos+1,closeParenPos-openParenPos-1),rmfNum) ) {
      std::ostringstream msg;
      msg << "Failed to get extver specifier from " << fileName << "\n";
      throw YellowAlert(msg.str());
    }
  }

  rmf inRMF;
  int status = inRMF.read(openFileName, rmfNum);
  if ( status != 0 ) 
  {
     std::ostringstream msg;
     msg << "\nFailed while reading from file " << openFileName <<"\n" << SPgetErrorStack() << "\n";
     throw YellowAlert(msg.str());
  }

  // Check for any warnings and write them out if chatter level is high enough
  string warnings =SPgetErrorStack();
  if ( warnings.length() > 0 && tpout.maxChatter() >= 15 ) tcout << warnings << std::endl;
  SPclearErrorStack();

  size_t nC = inRMF.NumberChannels();
  size_t nE = inRMF.NumberEnergyBins();

  // check for consistency with the spectrum number of detector channels
  if ((int)nC != spectrumDetchans) {
    std::ostringstream msg;
    msg << "Response DETCHANS: " << nC << ", and data DETCHANS: "
        << spectrumDetchans <<"\n   do not match.\n";
    throw YellowAlert(msg.str());
  }

  m_name = fileName;
  m_chanRangeIndicator = chanLimits;
  m_gqString = gqString;
  m_detectorChannels = nC;
  m_retainPosition = retainPositionArray;

  // set the response matrix information
  m_energyLow.resize(nE);
  m_energyHigh.resize(nE);
  m_binResponseGroups.resize(nE);
  for (size_t ie=0; ie<nE; ie++) {
    m_energyLow[ie] = inRMF.getLowEnergyElement(ie);
    m_energyHigh[ie] = inRMF.getHighEnergyElement(ie);
    m_binResponseGroups[ie] = inRMF.getNumberGroupsElement(ie);
  }

  // Prevent energy = 0 (or neg).
  bool smallEWarn = false;
  for (size_t i=0; i<nE; ++i) {
    if (m_energyHigh[i] < SMALL) {
      m_energyHigh[i] = SMALL;
      smallEWarn = true;
    }
    if (m_energyLow[i] < SMALL) {  
      m_energyLow[i] = std::max(SMALL, m_energyHigh[i]/1000.0);
      smallEWarn = true;
    }
  }
  if (smallEWarn) {
    tcout <<"***Warning: Detected response matrix energy bin value = 0 (or neg)." 
	  << "\n     XSPEC will instead use small finite value (response file will not be altered)."
	  << std::endl;
  }

  // load the response matrix information. 

  size_t offset = inRMF.getFirstChannel();

  m_binStart.resize(nE);
  m_binRunLengths.resize(nE);
  m_matrixValue.resize(nE);
  for (size_t ie=0; ie<nE; ie++) {
    size_t numGroups = inRMF.getNumberGroupsElement(ie);
    m_binStart[ie].resize(numGroups);
    m_binRunLengths[ie].resize(numGroups);
    size_t firstGroup = inRMF.getFirstGroupElement(ie);
    size_t numElements = 0;
    for (size_t k=0; k<numGroups; k++) {
      m_binStart[ie][k] = inRMF.getFirstChannelGroupElement(firstGroup+k) - offset;
      m_binRunLengths[ie][k] = inRMF.getNumberChannelsGroupElement(firstGroup+k);
      numElements += m_binRunLengths[ie][k];
    }
    m_matrixValue[ie].resize(numElements);
    size_t firstElement = inRMF.getFirstElementElement(firstGroup);
    for (size_t k=0; k<numElements; k++) {
      m_matrixValue[ie][k] = inRMF.getMatrixElement(firstElement+k);
    }
  }

  // normalize each row to unity and store the normalization factors

  normalizeRMF();

  // apply the current grouping

  groupRMF();

  // create the convolution arrays

  createConvolutionArray();

  // group and set the channel energies

  IntegerVector grouping;
  IntegerVector quality;
  decodeGQ(grouping, quality);

  bool qualitySet = !quality.empty();
  bool grouped = !grouping.empty();

  int startChan = m_chanRangeIndicator.startChan;
  int endChan = m_chanRangeIndicator.endChan;
  int firstChan = m_chanRangeIndicator.firstChan;
  size_t  ungroupedChannels = endChan - startChan + 1;
  size_t  groupedChannels=0;

  RealArray eMin(ungroupedChannels);
  RealArray eMax(ungroupedChannels);
  for (size_t i=0; i<ungroupedChannels; i++) {
    eMin[i] = inRMF.getChannelLowEnergyElement(i+(startChan-firstChan));
    eMax[i] = inRMF.getChannelHighEnergyElement(i+(startChan-firstChan));
  }

  // Prevent zero or neg energy from getting through.
  for (size_t i=0; i<ungroupedChannels; ++i) {
    if (eMin[i] < SMALL)  eMin[i] = SMALL;
    if (eMax[i] < SMALL)  eMax[i] = SMALL;
  }
  
  if (!grouped) {
    if (!qualitySet) {

      // Case I: no grouping or quality information.
      groupedChannels = ungroupedChannels;

    } else {

      // Case II: no grouping information, but quality information.
      for (size_t i=0; i<ungroupedChannels; i++) {
	if (quality[i] != 1) {
	  eMin[groupedChannels]   = eMin[i];
	  eMax[groupedChannels++] = eMax[i];
	}
      }
    }
    
  } else {

    // Case III General case: there is grouping information.

    size_t i=0;
    // Initialize eMin[0] and eMax[0] from first bin of OK quality,
    // regardless of whether grouping is = -1.  (A -1 for the first
    // bin is actually erroneous input.  This will cover for it.)
    while (i<ungroupedChannels) {
      if (!qualitySet || quality[i] != 1) {
	eMin[groupedChannels] = eMin[i];
	eMax[groupedChannels] = eMax[i];
	++i;
	++groupedChannels;
	break;
      }
      ++i;
    }

    if (eMin[0] < eMin[eMin.size()-1]) {
      while (i<ungroupedChannels) {
	if (!qualitySet || quality[i] != 1) {
	  if (grouping[i] != -1) {
	    eMin[groupedChannels] = eMin[i];
	    eMax[groupedChannels] = eMax[i];
	    ++i;
	    ++groupedChannels;
	  } else {
	    while (i<ungroupedChannels && grouping[i] == -1) {
	      if (!qualitySet || quality[i] != 1) {
		eMax[groupedChannels-1] = eMax[i];
	      }
	      ++i;  
	    }
	  }
	}
	else ++i;
      }
    } else { // Ebounds are in decreasing order
      while (i<ungroupedChannels) {
	if (!qualitySet || quality[i] != 1) {
	  if (grouping[i] != -1) {
	    eMin[groupedChannels] = eMin[i];
	    eMax[groupedChannels] = eMax[i];
	    ++i;
	    ++groupedChannels;
	  } else {
	    while (i<ungroupedChannels && grouping[i] == -1) {
	      if (!qualitySet || quality[i] != 1) {
		eMin[groupedChannels-1] = eMin[i];
	      }
	      ++i;  
	    }
	  }
	} else  ++i;
      }
     }
    
  }
  setEboundsMin(eMin[std::slice(0,groupedChannels,1)]);
  setEboundsMax(eMax[std::slice(0,groupedChannels,1)]);

  // this code was in setDescription

  m_telescope = inRMF.getTelescope();
  m_instrument = inRMF.getInstrument();
  m_channelType = inRMF.getChannelType();

  // there was code to read GSLOP_MIN(), GSLOP_MAX(), GOFFS_MIN() 
  // and GOFFS_MAX() keywords but we need to find another way to do this.

}




void ResponseMatrix::createConvolutionArray ()
{
   int nE = static_cast<int>(m_energyLow.size());
   RMFLONG lnE = static_cast<RMFLONG>(m_energyLow.size());
   std::vector<RMFLONG> positions;

   for (int i = 0; i < nE; i++)
   {
        int ngroups_i (m_binResponseGroups[i]);
        for (int j = 0; j<ngroups_i; j++)
        {
             int fchan_ij (m_binStart[i][j]);
	     int nchan_ij (m_binRunLengths[i][j]);
	     for (int k = 0; k <nchan_ij; k++)
                positions.push_back(lnE*(fchan_ij+k)+i);
        }
   }
   std::sort(positions.begin(), positions.end());

   size_t cvsz (positions.size());

   if (cvsz)
   {
        m_channelForRun.push_back(static_cast<int>(positions[0]/lnE)); //determine row
        m_energyStart.push_back(static_cast<int>(positions[0]%lnE));  //determine col
        m_energyRunLengths.push_back(1);
   }
   for (size_t i = 1; i<cvsz; i++)
   {
        RMFLONG pos_i (positions[i]);
        if (pos_i - positions[i-1] != 1 || pos_i % lnE == 0) //start of new group
        {
           m_channelForRun.push_back(static_cast<int>(pos_i/lnE)); 
           m_energyStart.push_back(static_cast<int>(pos_i%lnE));  
           m_energyRunLengths.push_back(1);
        }
        else
        {
           ++m_energyRunLengths.back();
        }
   }

   IntegerVector ecount(nE, 0);

   m_convArraySize = cvsz;

   m_convArray.reset(new Real[m_convArraySize]);
   size_t ncrg (m_channelForRun.size());
   size_t k(0);
   for (size_t i = 0; i < ncrg; i++)
   {
        int f_i (m_energyStart[i]);
        int n_i (m_energyRunLengths[i]);
        for (int j = 0; j < n_i; j++)
        {
             int ebin (f_i+j);
	     m_convArray[k++] = m_matrixValue[ebin][ecount[ebin]++];
        }
   }

   if (m_retainPosition) m_positions = positions;

   tpout << xsverbose(60);
   if (tpout.maxChatter() >= 60)
   {
        int  nchannels (0);
        tcout << "ORIGINAL\n";
        for (int i = 0; i < nE; i++)
        {
           for (int j = 0; j < m_binResponseGroups[i]; j++)
	   {
	      tcout << "Energy: " << i+1 << " nGroups: " << m_binResponseGroups[i] << 
	        " fchan: " << m_binStart[i][j] << " nchan: " << m_binRunLengths[i][j] << std::endl;
	      nchannels += m_binRunLengths[i][j];
           }

        }
        k = 0;
        for (size_t i = 0; i < m_matrixValue.size(); i++)
        {
	   for (size_t  j = 0; j < m_matrixValue[i].size(); j++)
	   {
	      tcout << "i: " << i << " j: " << j << " matrix: " 
                                << m_matrixValue[i][j] << std::endl;
              k++;
	   }
        }
        tcout << "total number of matrix elements: " << k << '\n';

        tcout << "\nFINAL\n";
        int  sz (m_channelForRun.size());

        for (int i = 0; i<sz; i++)
        {
	   tcout << "i: " << i << " channel: " 
                    << m_channelForRun[i] << " energyStart: " 
                    << m_energyStart[i] << " energyRunLen: " 
                    << m_energyRunLengths[i] << std::endl;
        }

        k = 0;
        RealArray exprow;
        size_t maxch (0);
        size_t numch (0);
        sz = m_energyRunLengths.size(); 
        bool okflag (1);
        tcout << "m_convArray.size: " << m_convArraySize << '\n';
        tcout << "i  " << "convArray   " << "matrix_ij" << '\n'; 
        for (int i = 0; i < sz; i++)
        {
           int  n_i (m_energyRunLengths[i]);
	   int  f_i (m_energyStart[i]);
	   int  m_i (m_channelForRun[i]);
           for (int j = 0; j < n_i; j++)
	   {
	      numch=0;
	      expandRmfRow(f_i+j,maxch,numch,exprow);
	      if (fabs(m_convArray[k] - exprow[m_i]) > 1.0e-15)
	      {
	              okflag = false;
	              tcout << "ERROR" <<std::endl;
	      }
	      tcout << k << "  " << m_convArray[k] << "  " 
                        << exprow[m_i] << std::endl;
	      k++;
	   }
        }
        if (okflag) tcout << "All elements match";
        else tcout << "There are errors";

        tcout << std::endl;

   }
   tpout << xsverbose();

   m_matrixValue.resize(0);
}

void ResponseMatrix::compressRmfRow (const size_t row, const size_t groupedResponseChannels, const RealArray& rmfRow)
{
  IntegerVector&   fchan=m_binStart[row];
  IntegerVector&   nchan=m_binRunLengths[row];
  RealArray&   matrixVal=m_matrixValue[row];

  size_t  ngroups=0;
  fchan.clear();
  nchan.clear();
  matrixVal.resize(0);

  size_t  rowSize(0);
  size_t  ch (0);
  size_t  startBin (0);

  while (ch < groupedResponseChannels)
  {
     while ((ch < groupedResponseChannels) && (rmfRow[ch] == 0))
        ch++;

     startBin = ch;   // index to the first non-zero bin

     while ((ch < groupedResponseChannels) && (rmfRow[ch] != 0))
        ch++;   // index to the last non-zero bin


     if (ch > startBin)
     {
        ngroups++;
	fchan.push_back(startBin);
	nchan.push_back(ch-startBin);		 
	rowSize += ch - startBin;
     }
  }

  binResponseGroups(row,ngroups);

  if (ngroups > 0)
  {
     matrixVal.resize(rowSize);
     size_t iCompressed = 0;
     for (size_t i=0; i<ngroups; ++i)
     {
        int nchan_i = nchan[i];
        int fchan_i = fchan[i];
        for (int j=0; j<nchan_i; ++j)
        {
           matrixVal[iCompressed] = rmfRow[fchan_i+j];
           ++iCompressed;
        }
     }
  }
}

void ResponseMatrix::expandRmfRow (const size_t row, size_t& maxChans, size_t& numChannels, RealArray &rmfRow)
{
  const size_t  ngroups=binResponseGroups(row);
  const IntegerVector&  fchan=binStart(row);
  const IntegerVector&  nchan=binRunLengths(row);
  const MatrixValue&   matrixVal=matrixValue();

  maxChans = fchan[ngroups-1] + nchan[ngroups-1];
  rmfRow.resize(maxChans, 0.0);

  for (size_t groupIndex=0; groupIndex<ngroups; groupIndex++)
  {
     RealArray temp = 
       matrixVal[row][std::slice(numChannels,nchan[groupIndex],1)];
     rmfRow[std::slice(fchan[groupIndex],nchan[groupIndex],1)] = temp;
     numChannels += nchan[groupIndex];
  }
}

void ResponseMatrix::normalizeRMF ()
{
  size_t  nE=energyLow().size();
  m_normFactor.resize(nE,0);
  for (size_t i=0; i<nE; ++i)
  {

     Real sum(0);
     if (m_matrixValue[i].size()) 
     {
         sum = m_matrixValue[i].sum();
         if (sum == 0.0)
         {
            // All entries in row are 0.0
            m_binResponseGroups[i] = 0;
            m_binStart[i].clear();
            m_binRunLengths[i].clear();
            m_matrixValue[i].resize(0);
         }
         else
         {
            m_matrixValue[i] /= sum;
            m_normFactor[i] = sum;
         }
     }
 }
}

void ResponseMatrix::groupRMF()
{
  IntegerVector grouping;
  IntegerVector quality;  

  decodeGQ(grouping, quality);

  int startChan = m_chanRangeIndicator.startChan - m_chanRangeIndicator.firstChan;
  int endChan = m_chanRangeIndicator.endChan - m_chanRangeIndicator.firstChan;
  int ungroupedSize = endChan - startChan + 1;

  std::unique_ptr<int[]> apChan2bin(new int[ungroupedSize]);
  int* chan2bin(apChan2bin.get());
  // chan2bin is maps chans (which may be offset by startChan) to
  // 0-based post-grouping bin number.  Its size should
  // always be ungroupedSize, which we'll assume is non-zero by
  // this point.
  int firstGood = 0;
  while (firstGood < ungroupedSize)
  {
     if (quality[firstGood] != 1) break;
     chan2bin[firstGood] = 0;
     ++firstGood;
  }
  if (firstGood == ungroupedSize) throw YellowAlert("All channels marked bad quality, cannot group RMF values.\n"); 
  int currBin = -1;
  for (int i=firstGood; i<ungroupedSize; ++i)
  {
     if (grouping[i] != -1 && quality[i] != 1)
     {
        ++currBin;
     }
     // currBin should never be -1 at this point.  If it is,
     // then something during patchGrouping func went wrong.
     if (currBin < 0) 
     {
        string msg("Patch grouping error detected in ResponseMatrixd::groupRMF");
        throw RedAlert(msg);
     }
     chan2bin[i] = currBin;
  } 
  int nGroupedBins = currBin + 1;

  int nE = m_energyLow.size();
  std::unique_ptr<Real[]> apGroupedRMFrow(new Real[nGroupedBins]);
  Real* groupedRMFrow(apGroupedRMFrow.get());
  MatrixIndex& allFChans = m_binStart;
  MatrixIndex& allNChans = m_binRunLengths;
  MatrixValue& allRmfVals = m_matrixValue;
  for (int iRow=0; iRow<nE; ++iRow)
  {
     int nGrp = binResponseGroups(iRow);
     if (nGrp)
     {
        IntegerVector& fChan = allFChans[iRow];
        IntegerVector& nChan = allNChans[iRow];
        RealArray& rmfVals = allRmfVals[iRow];
        // We know that grouping is never going to INCREASE the
        // channel numbers corresponding to non-zero matrix elements.
        // Therefore to save time, only bother to initialize the
        // groupedRMFrow elements that might possibly be used.
        int endWorkspace = std::max(fChan[nGrp-1] + nChan[nGrp-1] - (int)startChan, 0);
        endWorkspace = std::min(endWorkspace, nGroupedBins);
        // initial guess for startWorkspace:
        int startWorkspace = std::max(fChan[0] - (int)startChan, 0);
        // Now modify it due to grouping:
        if (startWorkspace >= ungroupedSize)
        {
           // No need for any groupedRMFrow workspace:
           startWorkspace = endWorkspace;
        }
        else
        {
           startWorkspace = chan2bin[startWorkspace];
        }
        for (int i=startWorkspace; i<endWorkspace; ++i) groupedRMFrow[i] = .0;
        // We also know that whatever funny stuff grouping might
        // do, increasing the total number of non-zero channels
        // isn't going to be one of 'em. 
        int sumOrgNChans = 0;
        for (int i=0; i<nGrp; ++i) sumOrgNChans += nChan[i];
        IntegerVector newElemIndices;
        newElemIndices.reserve(sumOrgNChans+1);
        // Make initial entry a dummy, simply so we know size
        // is always at least 1 and therefore don't have to
        // check in inner for loop.
        newElemIndices.push_back(-1);

        int iCompressed = 0;
        bool beyondSpec = false;
        for (int iGrp=0; iGrp<nGrp && !beyondSpec; ++iGrp)
        {
           int fChan_i = fChan[iGrp];
           int nChan_i = nChan[iGrp];
           for (int j=0; j<nChan_i; ++j)
           {
              int zeroBasedChan = fChan_i + j - startChan;
              if (zeroBasedChan >= ungroupedSize)
              {
                 beyondSpec = true;
              }
              else if (zeroBasedChan >= 0)
              {
                 if (quality[zeroBasedChan] != 1)
                 {
                    int newIndex = chan2bin[zeroBasedChan];
                    groupedRMFrow[newIndex] += rmfVals[iCompressed];
                    // newIndex is always being accessed in
                    // increasing order.
                    if (newIndex != newElemIndices.back())
                       newElemIndices.push_back(newIndex);
                 }
                 ++iCompressed;
              }
              else
              {
                 ++iCompressed;
              }
           }
        }

        size_t  ngroups=0;
        fChan.clear();
        nChan.clear();
        size_t  rowSize = newElemIndices.size() - 1;
        rmfVals.resize(rowSize);

        int prevChan = -999;
        int chanCount = 0;
        for (size_t i=1; i<=rowSize; ++i)
        {
           int currChan = newElemIndices[i];
           if (currChan != prevChan+1)
           {
              // start of new group, possibly the end
              // of a previous one.
              if (ngroups)  nChan.push_back(chanCount);
              ngroups++;
              fChan.push_back(currChan);
              chanCount = 0;              
           }
           rmfVals[i-1] = groupedRMFrow[currChan];
           prevChan = currChan;
           ++chanCount;
        }
        // Fill in nChan for last group (if there are any).
        if (chanCount) nChan.push_back(chanCount);

        binResponseGroups(iRow,ngroups);
     } // end if nGrp
  } // end nE loop

}

bool ResponseMatrix::gqMatch (const CodeContainer& value) const
{
  return  m_gqString==value;
}

void ResponseMatrix::decodeGQ (IntegerVector& group, IntegerVector& quality)
{
  DataUtility::decodeGQ(m_gqString, group, quality);
}

Real* ResponseMatrix::convArray ()
{
  return m_convArray.get();
}

Real ResponseMatrix::convArray (size_t index)
{
  return *(m_convArray.get()+index);
}

bool ResponseMatrix::compareChanRange (int firstChan, int startChan, int endChan) const
{
   return (firstChan == m_chanRangeIndicator.firstChan &&
      startChan == m_chanRangeIndicator.startChan &&
      endChan == m_chanRangeIndicator.endChan);
}

// Additional Declarations

// Handy function to check whether an input filename is a valid response.
// Also handles the cases of {n}, [n], or +n suffices. In the case of the 
// latter two the suffix will be changed to the appropriate {n} suffix.

bool isGoodResponseMatrixFile(string& fileName)
{
  // First need to check whether fileName has a {n}, [name,n], or +n suffix and 
  // split that off before calling the heasp routines
  string rmfName(fileName);
  int versNum;
  std::pair<string,int> extInfo;
  try {
    string tmpName(fileName);
    extInfo = XSparse::getSquareBracketContent(fileName, tmpName);
    versNum = XSparse::getCurlyBracketInt(tmpName, rmfName);
  } catch(...) {
    return false;
  }

  // Check whether fileName is a valid response file. uses heasp routines to
  // get number of EBOUNDS and MATRIX extensions. If extInfo.first has non-zero
  // length use that as the matrix extension name
  IntegerVector eboundsNums = RMFeboundsExtensions(rmfName);
  if ( eboundsNums.size() == 0 ) return false;
  IntegerVector matrixNums;
  if ( extInfo.first.length() == 0 ) {
    matrixNums = RMFmatrixExtensions(rmfName);
  } else {
    matrixNums = RMFmatrixExtensions(rmfName, extInfo.first);
  }
  if ( matrixNums.size() == 0 ) return false;

  // now do the special cases of [name,n], which means the nth instance of the
  // extension name, or [name], which means the first instance of the extension
  // name.

  if ( extInfo.first.length() != 0 ) {
    if ( extInfo.second == 0 ) {
      fileName = rmfName + "{1}";
    } else if ( extInfo.second < 0 || extInfo.second > (int)matrixNums.size() ) {
      // trap bad values of extInfo.second
      return false;
    } else {
      std::stringstream fileNameStream;
      fileNameStream << rmfName << "{" << extInfo.second << "}";
      fileName = fileNameStream.str();
    }
  }
      
  // now do the special cases of [n] or +n, which mean the nth extension.
  // need to check whether this is actually a MATRIX extension and assign it
  // to the correct instance of the extension

  if ( extInfo.first.length() == 0 && extInfo.second != 0 ) {
    int matrixVersion(0);
    for (size_t i=0; i<matrixNums.size(); i++) {
      if ( extInfo.second == matrixNums[i] ) matrixVersion = i+1;
    }
    if ( matrixVersion == 0 ) return false;
    std::stringstream fileNameStream;
    fileNameStream << rmfName << "{" << matrixVersion << "}";
    fileName = fileNameStream.str();
  }    

  // now check that any {n} suffix is valid

  versNum = XSparse::getCurlyBracketInt(fileName, rmfName);
  if ( versNum > (int)matrixNums.size() ) return false;

  // if we made it here then fileName is valid

  return true;
}

